package meshname

import (
	"errors"
	"net"
	"sync"
	"strings"

	"github.com/gologme/log"
	"github.com/miekg/dns"
)

type MeshnameServer struct {
	log          	*log.Logger
	listenAddr   	string
	dnsClient    	*dns.Client
	dnsServer    	*dns.Server
	networks     	map[string]*net.IPNet
	meshipNetworks	map[string]*net.IPNet
	enableMeshIP 	bool
	acmePort	string

	startedLock 	sync.RWMutex
	started     	bool
}

// New is a constructor for MeshnameServer
func New(log *log.Logger, listenAddr string, networks map[string]*net.IPNet, meshipNetworks map[string]*net.IPNet, enableMeshIP bool, acmePort string) *MeshnameServer {
	dnsClient := new(dns.Client)
	dnsClient.Timeout = 5000000000 // increased 5 seconds timeout

	return &MeshnameServer{
		log:          	log,
		listenAddr:   	listenAddr,
		networks:     	networks,
		meshipNetworks:	meshipNetworks,
		dnsClient:    	dnsClient,
		enableMeshIP: 	enableMeshIP,
		acmePort: 	acmePort,
	}
}

func (s *MeshnameServer) Stop() {
	s.startedLock.Lock()
	defer s.startedLock.Unlock()

	if s.started {
		if err := s.dnsServer.Shutdown(); err != nil {
			s.log.Debugln(err)
		}
		s.started = false
	}
}

func (s *MeshnameServer) Start() error {
	s.startedLock.Lock()
	defer s.startedLock.Unlock()

	if !s.started {
		waitStarted := make(chan struct{})
		s.dnsServer = &dns.Server{
			Addr:              s.listenAddr,
			Net:               "udp",
			NotifyStartedFunc: func() { close(waitStarted) },
		}
		if s.enableMeshIP {
			for mtld, subnet := range s.meshipNetworks {
				dns.HandleFunc(mtld, s.handleMeshIPRequest)
				s.log.Debugln("Handling as meship:", mtld, subnet)
			}
		}
		for tld, subnet := range s.networks {
			dns.HandleFunc(tld, s.handleMeshnameRequest)
			s.log.Debugln("Handling:", tld, subnet)
		}

		go func() {
			if err := s.dnsServer.ListenAndServe(); err != nil {
				s.log.Fatalln("MeshnameServer failed to start:", err)
			}
		}()
		<-waitStarted

		s.log.Debugln("MeshnameServer started")
		s.started = true
		return nil
	} else {
		return errors.New("MeshnameServer is already started")
	}
}

func (s *MeshnameServer) handleMeshnameRequest(w dns.ResponseWriter, r *dns.Msg) {
	var remoteLookups = make(map[string][]dns.Question)
	m := new(dns.Msg)
	m.SetReply(r)

	var tldList []string
	for tld := range s.networks {
		tldList = append(tldList, tld)
	}

	for _, q := range r.Question {
		labels := dns.SplitDomainName(q.Name)
		if len(labels) < 2 {
			s.log.Debugln("Error: invalid domain requested")
			continue
		}

		// process domain
		input := strings.ToLower(strings.TrimSuffix(q.Name, "."))
		subDomain := ""
		tld := ""
		for i := range tldList {
			t := strings.ToLower(tldList[i])
			if strings.HasSuffix(input, "."+t) {
				// define tld
				tld = t
				domainParts := strings.SplitN(input, "."+t, 2)
				firstPart := domainParts[0]
				firstPartLabels := dns.SplitDomainName(firstPart)
				// define hostname ("subDomain")
				subDomain = firstPartLabels[len(firstPartLabels)-1]
				if len(firstPartLabels) == 1 {
					s.handleMeshIPRequest(w, r)
					return
				}
			}
		}

		resolvedAddr, err := IPFromDomain(&subDomain)
		if err != nil {
			s.log.Debugln(err)
			continue
		}

		// check subnet validity
		if subnet, ok := s.networks[tld]; ok && subnet.Contains(resolvedAddr) {
			remoteLookups[resolvedAddr.String()] = append(remoteLookups[resolvedAddr.String()], q)
		} else {
			s.log.Debugln("Error: subnet doesn't match")
		}
	}

	for remoteServer, questions := range remoteLookups {
		rm := new(dns.Msg)
		rm.RecursionDesired = true
		rm.Question = questions
		// when specified, we add an alternative port for acme challenge requests
		portNumbers := []string{"53"}
		qdomain := rm.Question[0].Name
		dotIndex := strings.Index(qdomain, ".")
		qsubdomain := qdomain[:dotIndex]
		if rm.Question[0].Qtype == 16 && qsubdomain == "_acme-challenge" {
			s.log.Debugln("Is acme challenge. Trying on port "+s.acmePort+" too.")
			// add alternative port number
			portNumbers = append([]string{s.acmePort}, portNumbers...)
		}
		// make request
		err := error(nil)
		for _, port := range portNumbers {
			s.log.Debugln("trying port:"+port)
			resp, _, err := s.dnsClient.Exchange(rm, "["+remoteServer+"]:"+port)
			// if we had success we don't keep trying other ports
			if err == nil {
				s.log.Debugln(resp.String())
				m.Answer = append(m.Answer, resp.Answer...)
				m.Ns = append(m.Ns, resp.Ns...)
				m.Extra = append(m.Extra, resp.Extra...)
				break
			}
		}
		if err != nil {
			s.log.Debugln(err)
		}
	}

	if err := w.WriteMsg(m); err != nil {
		s.log.Debugln("Error writing response:", err)
	}
}

func (s *MeshnameServer) handleMeshIPRequest(w dns.ResponseWriter, r *dns.Msg) {
	m := new(dns.Msg)
	m.SetReply(r)
	m.Authoritative = true

	for _, q := range r.Question {
		labels := dns.SplitDomainName(q.Name)

/*		// handle A type properly
		if q.Qtype == dns.TypeA {
			answer := new(dns.SOA)
			answer.Hdr = dns.RR_Header{Name: q.Name, Rrtype: dns.TypeSOA, Class: dns.ClassINET, Ttl: 3600}
			answer.Ns = "ns1.mesh.kyun.li."
			answer.Mbox = "ns2.mesh.kyun.li."
			answer.Serial = 2024061000
			answer.Refresh = 1200
			answer.Retry = 180
			answer.Expire = 1209600
			answer.Minttl = 600
			m.SetEdns0(512, false)
			m.Ns = append(m.Ns, answer)
			s.log.Debugln(m)
			continue
		}
*/
		// resolve only AAAA type
		if q.Qtype != dns.TypeAAAA || q.Qclass != dns.ClassINET {
			//s.log.Debugln("Error: invalid resource requested")
			continue
		}

		if resolvedAddr, err := IPFromDomain(&labels[0]); err == nil {
			answer := new(dns.AAAA)
			answer.Hdr = dns.RR_Header{Name: q.Name, Rrtype: dns.TypeAAAA, Class: dns.ClassINET, Ttl: 3600}
			answer.AAAA = resolvedAddr
			m.Answer = append(m.Answer, answer)
			s.log.Debugln(m)
		} else {
			s.log.Debugln(err)
		}
	}

	if err := w.WriteMsg(m); err != nil {
		s.log.Debugln("Error writing response:", err)
	}
}

func (s *MeshnameServer) IsStarted() bool {
	s.startedLock.RLock()
	started := s.started
	s.startedLock.RUnlock()
	return started
}
